"""Efficient (O(log n) amortized update) Least-Recently-Used cache"""

import six

CacheMissError = KeyError

IsThreadsafe = False


class _node(object):
    __slots__ = ("prio", "index", "key", "value")

    def __init__(self, prio, index, key, value, int=int):
        self.prio = int(prio)
        self.index = int(index)
        self.key = key
        self.value = value

    def __richcmp__(self, other, op):
        if op == 0:
            return self.prio < other.prio
        elif op == 1:
            return self.prio <= other.prio
        elif op == 2:
            return self.prio == other.prio
        elif op == 3:
            return self.prio != other.prio
        elif op == 4:
            return self.prio > other.prio
        elif op == 5:
            return self.prio >= other.prio
        else:
            return False


class LRUCache(object):
    """Least-Recently-Used (LRU) cache.

    Instances of this class behave like a Python mapping type, except that
    its size is guaranteed to never surpass the assigned limit and when
    spillover occurs, the least recently used items get removed first.
    """

    def __init__(self, size, touch_on_read=True, eviction_callback=None):
        self.size = size
        self.touch_on_read = touch_on_read
        self.pqueue = []
        self.emap = {}
        self.next_prio = 0
        self.eviction_callback = eviction_callback

    def __len__(self):
        return len(self.pqueue)

    def __contains__(self, key):
        return key in self.emap

    def rehash(self):
        bprio = self.pqueue[0].prio
        for i in xrange(len(self.pqueue)):
            node = self.pqueue[i]
            node.prio -= bprio
        self.next_prio -= bprio

    def decrease(self, node):
        node.prio = self.next_prio
        self.next_prio += 1

        if self.next_prio > 0xFFFFFFF0:
            self.rehash()

        pqueue = self.pqueue
        sz = len(pqueue)
        while 1:
            ix = node.index
            l = 2 * ix + 1
            r = 2 * ix + 2

            if r < sz:
                ln = pqueue[l]
                rn = pqueue[r]

            if r < sz and rn.prio < ln.prio:
                sw = r
                swn = rn
            elif l < sz:
                sw = l
                swn = pqueue[l]
            else:
                break

            pqueue[sw] = node
            pqueue[ix] = swn
            node.index = sw
            swn.index = ix

    def iterkeys(self):
        return self.emap.iterkeys()

    def itervalues(self):
        return LRUCacheValuesIterator(self)

    def iteritems(self):
        return LRUCacheItemsIterator(self)

    def keys(self):
        return self.emap.keys()

    def values(self):
        return list(self.itervalues())

    def items(self):
        return list(self.iteritems())

    def __iter__(self):
        return self.iterkeys()

    def __setitem__(self, key, val):
        if key in self.emap:
            node = self.emap[key]
            node.value = val
            self.decrease(node)
        elif len(self.pqueue) >= self.size:
            node = self.pqueue[0]
            oldkey = node.key
            oldval = node.value
            del self.emap[node.key]
            node.key = key
            node.value = val
            self.emap[key] = node
            self.decrease(node)
            if self.eviction_callback is not None:
                self.eviction_callback(oldkey, oldval)
        else:
            node = _node(self.next_prio, len(self.pqueue), key, val)
            self.emap[key] = node
            self.pqueue.append(node)
            self.next_prio += 1
            if self.next_prio > 0xFFFFFFF0:
                self.c_rehash()

    def __getitem__(self, key):
        if key not in self.emap:
            raise CacheMissError(key)
        else:
            node = self.emap[key]
            if self.touch_on_read:
                self.decrease(node)
            return node.value

    def __delitem__(self, key):
        if key not in self.emap:
            raise CacheMissError(key)
        else:
            node = self.emap[key]
            self.decrease(node)

            node2 = self.pqueue[-1]
            if node2 is not node:
                self.pqueue[node.index] = node2
                node2.index = node.index

            del self.emap[key]
            del self.pqueue[-1]

    def cas(self, key, oldvalue, newvalue):
        if key in self.emap:
            node = self.emap[key]
            if node.value is oldvalue:
                node.value = newvalue
                self.decrease(node)
                return True
            elif self.touch_on_read:
                self.decrease(node)
        return False

    def get(self, key, deflt=None):
        if key not in self.emap:
            return deflt
        else:
            node = self.emap[key]
            self.decrease(node)
            return node.value

    def pop(self, key, deflt=CacheMissError):
        if key not in self.emap:
            if deflt is CacheMissError:
                raise CacheMissError(key)
            else:
                rv = deflt
        else:
            rv = self[key]
            del self[key]
        return rv

    def setdefault(self, key, deflt=None):
        if key not in self.emap:
            self[key] = deflt
            return deflt
        else:
            node = self.emap[key]
            self.decrease(node)
            return node.value

    def update(self, iterOrDict):
        if isinstance(iterOrDict, dict) or isinstance(iterOrDict, LRUCache):
            for k, v in six.iteritems(iterOrDict):
                self[k] = v
        else:
            for k, v in iterOrDict:
                self[k] = v

    def clear(self):
        self.pqueue = []
        self.emap = {}
        self.next_prio = 0

    def defrag(self):
        self.pqueue = list(self.pqueue)
        self.emap = self.emap.copy()

    def __repr__(self):
        return "<LRUCache (%d elements, %d max)>" % (len(self), self.size)


class LRUCacheItemsIterator(object):
    __slots__ = ("cache", "pos")

    def __init__(self, cache):
        self.cache = cache
        self.pos = 0

    def __next__(self):
        if self.pos >= len(self.cache.pqueue):
            raise StopIteration
        else:
            node = self.cache.pqueue[self.pos]
            self.pos += 1
            return (node.key, node.value)

    def __iter__(self):
        return self

    next = __next__


class LRUCacheValuesIterator(object):
    __slots__ = ("cache", "pos")

    def __init__(self, cache):
        self.cache = cache
        self.pos = 0

    def __next__(self):
        if self.pos >= len(self.cache.pqueue):
            raise StopIteration
        else:
            node = self.cache.pqueue[self.pos]
            self.pos += 1
            return node.value

    def __iter__(self):
        return self

    next = __next__



cdef struct _node

from cpython.object cimport PyObject

cdef class LazyCuckooCache:
    cdef object __weakref__

    cdef unsigned long long _nextprio

    cdef bint _rnd_data[16]
    cdef unsigned char _rnd_pos

    cdef object eviction_callback
    cdef object _hash2seed
    cdef object hash1
    cdef object hash2

    cdef _node *table

    cdef bint _rehash_in_progress

    cdef readonly unsigned int size
    cdef readonly unsigned int initial_size
    cdef unsigned int table_size
    cdef readonly unsigned int nitems

    cdef readonly bint touch_on_read

    cdef unsigned int _hash1(LazyCuckooCache self, x) except? 0xFFFFFFFF
    cdef unsigned int _hash2(LazyCuckooCache self, x) except? 0xFFFFFFFF
    cdef unsigned long long _assign_prio(LazyCuckooCache self)

    cdef int _add_node(LazyCuckooCache self, _node *table, unsigned int tsize, _node *node,
            unsigned int h1, unsigned int h2, PyObject *key, PyObject *value,
            unsigned long long prio, unsigned int item_diff, bint recheck) except -1
    cdef int _rehash(LazyCuckooCache self) except -1

    cdef int _init_rnd(self) except -1
    cdef bint _rnd(LazyCuckooCache self)

    cpdef pop(LazyCuckooCache self, key, deflt = ?)
    cpdef setdefault(self, key, deflt = ?)
    cpdef get(self, key, deflt = ?)




cdef class _node:
    # private, do not use
    cdef unsigned int prio
    cdef unsigned int index
    cdef object key
    cdef object value

    cdef attach(_node self, unsigned int prio, unsigned int index, object key, object value)
    cdef detach(_node self)

cdef class LRUCache:
    cdef int use_freelist
    cdef unsigned int next_prio
    cdef readonly unsigned int size
    cdef readonly unsigned int touch_on_read
    cdef list pqueue
    cdef list freelist
    cdef dict emap
    cdef object eviction_callback

    cdef void c_rehash(LRUCache self)
    cdef void c_decrease(LRUCache self, _node node) # private, do not use
    cdef int c__setitem__(LRUCache self, object key, object val) except -1
    cdef object c__getitem__(LRUCache self, key)
    cdef int c__delitem__(LRUCache self, key) except -1
    cdef c_get(LRUCache self, object key, object deflt)



# -*- coding: utf-8 -*-
"""Efficient (O(1) amortized update) Quasi-Least-Recently-Used cache"""
from random import random
import functools

import six
from six.moves import range as xrange

CacheMissError = KeyError

IsThreadsafe = False


class _node(object):
    __slots__ = ("key", "value", "h1", "h2", "prio")

    def __init__(self, key, value, h1, h2, prio):
        self.key = key
        self.value = value
        self.h1 = h1
        self.h2 = h2
        self.prio = prio


def _hash2(seed, x):
    return hash((seed, x))


class LazyCuckooCache(object):
    """Quasi-Least-Recently-Used (LRU) cache.

    Instances of this class behave like a Python mapping type, except that
    its size is guaranteed to never surpass the assigned limit and when
    spillover occurs, an item likely not recently used gets removed first.

    The algorithm is O(1) worst case for lookup, and amortized O(1) for
    update. Update can only incur more than O(1) if rehashing needs to take
    place, something that can be avoided by passing preallocate = True

    The default hash functions may not be at all optimal, so you're encouraged
    to provide your own if you know the type of keys you'll be hashing.
    """

    # fmt: off
    def __init__(self, size, touch_on_read=True, eviction_callback=None,
            preallocate=True, hash1=None, hash2=None, initial_size=256):
        # fmt: on
        if size <= 0:
            raise ValueError("Cannot build a size-0 cache")

        self.size = size
        self.touch_on_read = touch_on_read
        if preallocate:
            initial_size = size
        else:
            initial_size = min(initial_size, size)
        initial_size = max(1, initial_size)
        self.initial_size = initial_size
        self.table = [None] * initial_size
        self.nitems = 0
        self.eviction_callback = eviction_callback
        if hash1 is None:
            hash1 = hash
        self.hash1 = hash1
        if hash2 is None:
            self.hash2 = functools.partial(_hash2, random() * 0.98 + 0.01)
        else:
            self.hash2 = hash2
        self._nextprio = 0

    def _assign_prio(self):
        prio = self._nextprio
        self._nextprio += 1
        if prio >= 0x7FFFFFFFFFFFFFFE:
            # The garbage overfloweth, repriorize
            # by truncating the lower half of the priority range
            # shouldn't happen not even once in a blue moon
            for node in self.table:
                if node is not None:
                    nprio = node.prio
                    if nprio > 0x4000000000000000:
                        node.prio = nprio - 0x4000000000000000
                    else:
                        node.prio = 0
            self._nextprio -= 0x4000000000000000
            prio -= 0x4000000000000000
        return prio

    def __len__(self):
        return self.nitems

    def __contains__(self, key):
        table = self.table
        tsize = len(table)
        h1 = self.hash1(key)
        node = table[h1 % tsize]
        if node is not None and node.key == key:
            return True

        h2 = self.hash2(key)
        node = table[h2 % tsize]
        return node is not None and node.key == key

    def rehash(self):
        otable = self.table
        size = self.size
        ntablesize = len(otable)
        if ntablesize >= size:
            return False

        ntablesize += max(1, ntablesize / 2)
        if ntablesize > size:
            ntablesize = size

        ntable = [None] * ntablesize
        _add_node = self._add_node
        eviction_callback = self.eviction_callback

        # Some evictions might take place during rehashing
        nitems = 0
        for node in otable:
            if node is not None:
                if not _add_node(ntable, ntablesize, node, node.h1, node.h2, node.key, node.value, node.prio, 0):
                    if eviction_callback is not None:
                        eviction_callback(node.key, node.value)
                else:
                    nitems += 1
        self.table = ntable
        self.nitems = nitems

        return True

    def _add_node(self, table, tsize, node, h1, h2, key, value, prio, item_diff):
        ix1 = h1 % tsize
        tnode = table[ix1]
        if tnode is not None:
            if tnode.key == key:
                # Replacing key, no eviction callback
                if node is None:
                    tnode.value = value
                    tnode.prio = prio
                else:
                    table[ix1] = node
                return True
        else:
            # Free slot
            if node is None:
                node = _node(key, value, h1, h2, prio)
            if item_diff:
                self.nitems += item_diff
            table[ix1] = node
            return True

        ix2 = h2 % tsize
        tnode = table[ix2]
        if tnode is not None:
            if tnode.key == key:
                # Replacing key, no eviction callback
                if node is None:
                    tnode.value = value
                    tnode.prio = prio
                else:
                    table[ix2] = node
                return True
        else:
            # Free slot
            if node is None:
                node = _node(key, value, h1, h2, prio)
            if item_diff:
                self.nitems += item_diff
            table[ix2] = node
            return True

        # No room
        return False

    def iterkeys(self):
        for node in self.table:
            if node is not None:
                yield node.key

    def itervalues(self):
        for node in self.table:
            if node is not None:
                yield node.value

    def iteritems(self):
        for node in self.table:
            if node is not None:
                yield (node.key, node.value)

    def keys(self):
        return list(self.iterkeys())

    def values(self):
        return list(self.itervalues())

    def items(self):
        return list(self.iteritems())

    def __iter__(self):
        return self.iterkeys()

    def __setitem__(self, key, value):
        h1 = self.hash1(key)
        h2 = self.hash2(key)
        table = self.table
        tsize = len(table)
        prio = self._assign_prio()
        _add_node = self._add_node
        if not _add_node(table, tsize, None, h1, h2, key, value, prio, 1):
            if self.rehash():
                # Try agin
                table = self.table
                tsize = len(table)
                if _add_node(table, tsize, None, h1, h2, key, value, prio, 1):
                    return

            # No room, evict some entry, pick one of the two options randomly
            tnode1 = table[h1 % tsize]
            tnode2 = table[h2 % tsize]
            prio1 = tnode1.prio
            prio2 = tnode2.prio
            if prio1 < prio1 or (prio1 == prio2 and random() < 0.5):
                tnode = tnode1
            else:
                tnode = tnode2
            eviction_callback = self.eviction_callback
            k = tnode.key
            v = tnode.value
            tnode.key = key
            tnode.value = value
            tnode.h1 = h1
            tnode.h2 = h2
            tnode.prio = prio
            if eviction_callback is not None:
                eviction_callback(k, v)

    def __getitem__(self, key):
        table = self.table
        tsize = len(table)
        h1 = self.hash1(key)
        node = table[h1 % tsize]
        if node is not None and node.key == key:
            if self.touch_on_read:
                node.prio = self._assign_prio()
            return node.value

        h2 = self.hash2(key)
        node = table[h2 % tsize]
        if node is not None and node.key == key:
            if self.touch_on_read:
                node.prio = self._assign_prio()
            return node.value

        raise CacheMissError(key)

    def __delitem__(self, key):
        return self.pop(key)

    def cas(self, key, oldvalue, newvalue):
        table = self.table
        tsize = len(table)
        h1 = self.hash1(key)
        node = table[h1 % tsize]
        if node is not None and node.key == key:
            if node.value is oldvalue:
                node.value = newvalue
                node.prio = self._assign_prio()
                return True
            else:
                return False

        h2 = self.hash2(key)
        node = table[h2 % tsize]
        if node is not None and node.key == key:
            if node.value is oldvalue:
                node.value = newvalue
                node.prio = self._assign_prio()
                return True
            else:
                return False

        return False

    def get(self, key, deflt=None):
        table = self.table
        tsize = len(table)
        h1 = self.hash1(key)
        node = table[h1 % tsize]
        if node is not None and node.key == key:
            if self.touch_on_read:
                node.prio = self._assign_prio()
            return node.value

        h2 = self.hash2(key)
        node = table[h2 % tsize]
        if node is not None and node.key == key:
            if self.touch_on_read:
                node.prio = self._assign_prio()
            return node.value

        return deflt

    def pop(self, key, deflt=CacheMissError):
        table = self.table
        tsize = len(table)
        h1 = self.hash1(key)
        ix1 = h1 % tsize
        node = table[ix1]
        if node is not None and node.key == key:
            table[ix1] = None
            self.nitems -= 1
            return node.value

        h2 = self.hash2(key)
        ix2 = h2 % tsize
        node = table[ix2]
        if node is not None and node.key == key:
            table[ix2] = None
            self.nitems -= 1
            return node.value

        if deflt is CacheMissError:
            raise CacheMissError(key)
        else:
            return deflt

    def setdefault(self, key, deflt=None):
        for j in xrange(3):
            table = self.table
            tsize = len(table)
            h1 = self.hash1(key)
            ix1 = h1 % tsize
            node1 = table[ix1]
            if node1 is not None and node1.key == key:
                if self.touch_on_read:
                    node1.prio = self._assign_prio()
                return node1.value

            h2 = self.hash2(key)
            ix2 = h2 % tsize
            node2 = table[ix2]
            if node2 is not None and node2.key == key:
                if self.touch_on_read:
                    node2.prio = self._assign_prio()
                return node2.value

            prio = self._assign_prio()
            if node1 is None:
                table[ix1] = _node(key, deflt, h1, h2, prio)
                self.nitems += 1
            elif node2 is None:
                table[ix2] = _node(key, deflt, h1, h2, prio)
                self.nitems += 1
            elif j < 2 and self.rehash():
                # Enlarged the table, retry
                continue
            else:
                # Pick a node to evict
                if random() < 0.5:
                    tnode = node1
                else:
                    tnode = node2
                eviction_callback = self.eviction_callback
                if eviction_callback is not None:
                    eviction_callback(tnode.key, tnode.value)
                tnode.key = key
                tnode.value = deflt
                tnode.h1 = h1
                tnode.h2 = h2
                tnode.prio = prio
            return deflt
        else:
            # Unreachable code, in theory
            raise AssertionError("unreachable")

    def update(self, iterOrDict):
        if self is iterOrDict:
            return
        if isinstance(iterOrDict, dict) or isinstance(iterOrDict, LazyCuckooCache):
            for k, v in six.iteritems(iterOrDict):
                self[k] = v
        else:
            for k, v in iterOrDict:
                self[k] = v

    def clear(self):
        self.table = [None] * self.initial_size
        self.nitems = 0

    def defrag(self):
        self.table = list(self.table)

    def __repr__(self):
        return "<LazyCuckooCache (%d elements, %d max)>" % (len(self), self.size)



# -*- coding: utf-8 -*-
"""Efficient (O(1) amortized update) Quasi-Least-Recently-Used cache

This implementation provides atomic operations for all single-element operations.
This includes get, set, delete, setdefault and cas.

Additionally, keys, values and items will provide an atomic snapshot of
some consistent state of the cache.
"""
cimport cython

from libc.stdlib cimport malloc, free
from cpython.ref cimport Py_CLEAR, Py_XINCREF, Py_XDECREF, Py_INCREF
from cpython.object cimport (
    PyObject_RichCompareBool, Py_EQ, PyObject, visitproc, traverseproc, inquiry, PyTypeObject)
from cpython.list cimport PyList_New

cdef extern from "Python.h":
    # Similar to cpython.list's version, but it takes a PyObject*
    # Allows us to avoid generating useless (and risky) Py_DECREF calls
    # due to reference count handling of regular <object> variables
    void PyList_SET_ITEM(object list, Py_ssize_t i, PyObject *o)

IsThreadsafe = True

from random import random
import functools

CacheMissError = KeyError
cdef object CacheMissError_ = CacheMissError

# Hash table manipulation helpers
#
# Hash nodes, when empty, will contain NULL in both the key and value
# While the code doesn't assume key is NULL iff value is NULL, it ought to be true
# at all times. Still, it's safer not to assume it.
#
# In theory, tables could be rebuilt in-place. In practice, however, the process
# is difficult, and impossible to do atomically if concurrent activity from other
# threads is allowed. So we just rebuild them in a new array that we swap with the
# old one afterwards.
#
# Hashes are 32-bit, the implementation doesn't support tables bigger than 4G items
# in any case, and doing so, while simple, would involve making nodes quite bigger,
# which doesn't seem worthwhile.
#
# Nodes contain both hashes for their key, if non-empty. While the utility of the
# hashes themselves isn't huge, and could be avoided and recomputed (they're only
# ever needed while growing the table), they help make growing the hash table
# not only faster, but also atomic. Without the hashes, we'd have to invoke
# abritrary user code to recompute them, which would break atomicity of the
# rehash procedure, and cause a lot of trouble. Having the hashes, thus, is a
# small price to pay for an atomic, lockless yet thread-safe implementation.

cdef struct _node:
    PyObject *key
    PyObject *value
    unsigned long long prio
    unsigned int h1
    unsigned int h2

cdef int _init_table_items(_node* table, unsigned int start, unsigned int end) except -1:
    cdef unsigned int i
    for i from start <= i < end:
        table[i].key = NULL
        table[i].value = NULL
        table[i].prio = 0
        table[i].h1 = 0
        table[i].h2 = 0
    return 0

cdef int _free_table_items(_node* table, unsigned int start, unsigned int end) except -1:
    cdef unsigned int i
    for i from start <= i < end:
        Py_CLEAR(table[i].key)
        Py_CLEAR(table[i].value)
    return 0

cdef _node* _alloc_table(unsigned int size) except NULL:
    cdef _node* table

    assert size > 0

    table = <_node*>malloc(sizeof(_node) * size)
    if table == NULL:
        raise MemoryError

    _init_table_items(table, 0, size)
    return table

cdef int _free_table(_node* table, unsigned int size) except -1:
    _free_table_items(table, 0, size)
    free(table)
    return 0

cdef int _key_equals(_node* node, object key, unsigned int h1) except -1:
    cdef PyObject *nkey
    if node != NULL and node.key != NULL:
        if node.key == <PyObject*>key:
            return 1
        elif node.h1 != h1:
            return 0
        nkey = node.key
        if PyObject_RichCompareBool(<object>nkey, key, Py_EQ):
            return 1
    return 0

cdef int _value_set(_node* node, PyObject *value, unsigned long long prio) except -1:
    cdef PyObject *t
    t = node.value

    # Must keep reference decrefs at the end to avoid invoking python code
    # with inconsistent node states
    Py_XINCREF(value)
    node.value = value
    node.prio = prio
    Py_XDECREF(t)
    return 0

cdef int _node_set(_node* node, PyObject *key, PyObject *value,
        unsigned int h1, unsigned int h2, unsigned long long prio) except -1:
    cdef PyObject *tk
    cdef PyObject *tv
    tk = node.key
    tv = node.value

    # Must keep reference decrefs at the end to avoid invoking python code
    # with inconsistent node states
    Py_XINCREF(key)
    Py_XINCREF(value)
    node.key = key
    node.value = value
    node.prio = prio
    node.h1 = h1
    node.h2 = h2
    Py_XDECREF(tk)
    Py_XDECREF(tv)
    return 0

cdef class LazyCuckooCache:
    """Quasi-Least-Recently-Used (LRU) cache.

    Instances of this class behave like a Python mapping type, except that
    its size is guaranteed to never surpass the assigned limit and when
    spillover occurs, an item likely not recently used gets removed first.

    The algorithm is O(1) worst case for lookup, and amortized O(1) for
    update. Update can only incur more than O(1) if rehashing needs to take
    place, something that can be avoided by passing preallocate = True

    The default hash functions may not be at all optimal, so you're encouraged
    to provide your own if you know the type of keys you'll be hashing.
    """

    #
    # Atomicity considerations
    #
    # Each call to hash1, hash2 or key_equals will break atomicity. Thus, it is necessary
    # to consider them as atomic barriers, and assume the hash table could have changed
    # considerably across those calls. All pointers to nodes to the old table could be
    # invalid.
    #
    # To cope with that, when possible, the calls should be the very last thing done,
    # after grabbing all the information required to perform whichever operation from
    # the relevant nodes.
    #
    # When manipulation of the table is the operation to be performed, after key_equals
    # table mutation needs to be assertained. A recheck of the table pointer against
    # the local copy, and the key's object identity should suffice in most contexts
    # to be able to proceed with the operation safely, since the table never grows,
    # and operations are always local to the node.
    #
    # If concurrent mutation happens at the atomic barrier, however, the operation needs
    # to be restarted. So, no change of the hash table must be applied until after
    # the atomic barrier, to avoid leaving an inconsistent state. Sometimes, such a
    # restart isn't necessary, like when updating node hit counters (prio). Failure
    # to increment a node's hit counter isn't relevant, so a better approach in those
    # cases is to just skip it.
    #
    # Each time there's a conflict, at least one of the conflicting operation succeeds.
    # This guarantees the application will be making progress as a whole (though a thread
    # in particular might not). This is a concept akin to software-transactional memory.

    def __cinit__(self, unsigned int size, bint touch_on_read = True, eviction_callback = None,
            bint preallocate = True, hash1 = None, hash2 = None, unsigned int initial_size = 256):
        cdef _node *table

        if size <= 0:
            raise ValueError("Cannot build a size-0 cache")

        self.table = NULL
        self.size = size
        self._rehash_in_progress = False
        self.touch_on_read = touch_on_read
        if preallocate:
            initial_size = size
        else:
            initial_size = min(initial_size, size)
        initial_size = max(1, initial_size)
        self.initial_size = initial_size
        self.nitems = 0
        self.eviction_callback = eviction_callback
        self.hash1 = hash1
        self.hash2 = hash2
        if hash2 is None:
            self._hash2seed = random() * 0.98 + 0.01
        self._nextprio = 0
        self.table_size = initial_size
        self.table = _alloc_table(initial_size)

        self._init_rnd()

    def __dealloc__(self):
        if self.table != NULL and self.table_size > 0:
            table = self.table
            table_size = self.table_size
            self.table = NULL
            self.table_size = 0
            _free_table(table, table_size)

    cdef unsigned int _hash1(self, x) except? 0xFFFFFFFF:
        if self.hash1 is None:
            return <unsigned int><long>hash(x)
        else:
            return <unsigned int><long>self.hash1(x)

    cdef unsigned int _hash2(self, x) except? 0xFFFFFFFF:
        if self.hash2 is None:
            # Replicates what tuple's hash does with (hash2seed, x)
            return <unsigned int>((0x345678L ^ <long>hash(x)) * 1000003L + 97531L)
        else:
            return <unsigned int><long>self.hash2(x)

    cdef unsigned long long _assign_prio(self):
        cdef unsigned int i, tsize
        cdef unsigned long long prio, nprio
        cdef _node *table

        prio = self._nextprio
        self._nextprio += 1
        if prio >= 0x7FFFFFFFFFFFFFFEULL:
            # The garbage overfloweth, repriorize
            # by truncating the lower half of the priority range
            # shouldn't happen not even once in a blue moon
            # This whole loop should be atomic (albeit perhaps not fast)
            table = self.table
            tsize = self.table_size
            for i from 0 <= i < tsize:
                node = table+i
                if node.key != NULL:
                    nprio = node.prio
                    if nprio > 0x4000000000000000ULL:
                        node.prio = nprio - 0x4000000000000000ULL
                    else:
                        node.prio = 0
            self._nextprio -= 0x4000000000000000ULL
            prio -= 0x4000000000000000ULL
        return prio

    def __len__(self):
        return self.nitems

    @cython.cdivision(True)
    cdef int _init_rnd(self) except -1:
        self._rnd_pos = 0
        for i from 0 <= i < (sizeof(self._rnd_data) / sizeof(self._rnd_data[0])):
            self._rnd_data[i] = random() < 0.5
        return 0

    @cython.cdivision(True)
    cdef bint _rnd(self):
        # This "maybe random" helper gives us a random-ish value without invoking arbitrary
        # python code, and thus break atomicity. Perfect randomness isn't necessary, but thread-safety is.
        rpos = self._rnd_pos
        self._rnd_pos = (self._rnd_pos + 1) % (sizeof(self._rnd_data) / sizeof(self._rnd_data[0]))
        return self._rnd_data[rpos]

    @cython.cdivision(True)
    def __contains__(LazyCuckooCache self, key):
        cdef _node *node
        cdef unsigned int tsize, h1, h2

        h1 = self._hash1(key)
        table = self.table
        tsize = self.table_size
        node = table + (h1 % tsize)
        if _key_equals(node, key, h1):
            return True

        h2 = self._hash2(key)
        table = self.table
        tsize = self.table_size
        node = table + (h2 % tsize)
        return _key_equals(node, key, h1)

    cdef int _rehash(LazyCuckooCache self) except -1:
        cdef unsigned int i, tsize, ntablesize, size, nitems

        otable = self.table
        size = self.size
        tsize = ntablesize = self.table_size
        if ntablesize >= size or self._rehash_in_progress:
            return 0

        eviction_callback = self.eviction_callback
        if eviction_callback is not None:
            # Atomic barrier, so recheck table afterwards
            evict_keys = []
            evict_values = []
            otable = self.table
            tsize = ntablesize = self.table_size
            if ntablesize >= size or self._rehash_in_progress:
                return 0

        self._rehash_in_progress = True
        try:
            ntablesize += max(1, ntablesize / 2)
            if ntablesize > size:
                ntablesize = size

            ntable = _alloc_table(ntablesize)
            try:
                _init_table_items(ntable, 0, ntablesize)

                # Some evictions might take place during rehashing
                nitems = 0
                for i from 0 <= i < tsize:
                    node = otable + i
                    if node.key != NULL and node.value != NULL:
                        if not self._add_node(ntable, ntablesize, node, node.h1, node.h2,
                                node.key, node.value, node.prio, 0, False):
                            # Queue eviction notification, not an atomic barrier since we just queue it
                            # without creating new objects, so no GC and no arbitrary python code can be triggered
                            if eviction_callback is not None:
                                k = <object>node.key
                                v = <object>node.value
                                evict_keys.append(k)
                                evict_values.append(v)
                                del k, v
                        else:
                            nitems += 1
            except:
                _free_table(ntable, ntablesize)
                raise

            self.table = ntable
            self.table_size = ntablesize
            self.nitems = nitems
            _free_table(otable, tsize)
        finally:
            self._rehash_in_progress = False

        if eviction_callback is not None:
            num_evictions = len(evict_keys)
            for i from 0 <= i < num_evictions:
                eviction_callback(evict_keys[i], evict_values[i])

        return 1

    @cython.cdivision(True)
    cdef int _add_node(self, _node *table, unsigned int tsize, _node *node, unsigned int h1, unsigned int h2,
            PyObject *key, PyObject *value, unsigned long long prio, unsigned int item_diff, bint recheck) except -1:
        cdef PyObject *tkey

        if recheck:
            # Recheck-enabled add_node should always receive the instance's table at entry
            assert self.table == table
            assert self.table_size == tsize

        while True:
            if recheck:
                table = self.table
                tsize = self.table_size

            # Adds a node `node` to the table `table`. The node's attributes h1, h2, key and value must
            # always be specified, but the node itself may be omitted.
            #
            # If the table is the instance's table, and can thus receive concurrent modifications
            # at atomic barriers, recheck must be given as True, and the operation will be retried on conflicts.
            #
            # If the table is a local copy, and the operation is a rebuild, recheck must be given as False to
            # avoid equality checks, since during rebuild, equal yet different keys cannot be enocuntered. An
            # identity check will be used instead, since within a hash table, equal keys are identical keys.
            tnode = table + (h1 % tsize)
            if tnode.key != NULL:
                tkey = tnode.key
                
                if (recheck and _key_equals(tnode, <object>key, h1)) or (not recheck and tkey == key):
                    if recheck and (table != self.table or tkey != tnode.key):
                        # Conflict during equals, restart
                        continue
    
                    # Replacing key, no eviction callback
                    if node == NULL:
                        _value_set(tnode, value, prio)
                    else:
                        _node_set(tnode, key, value, h1, h2, prio)
                    return 1
            else:
                # Free slot
                if item_diff:
                    self.nitems += item_diff
                _node_set(tnode, key, value, h1, h2, prio)
                return 1

            tnode = table + (h2 % tsize)
            if tnode.key != NULL:
                tkey = tnode.key
                if (recheck and _key_equals(tnode, <object>key, h1)) or (not recheck and tkey == key):
                    if recheck and (table != self.table or tkey != tnode.key):
                        # Conflict during equals, restart
                        continue
                    # Replacing key, no eviction callback
                    if node == NULL:
                        _value_set(tnode, value, prio)
                    else:
                        _node_set(tnode, key, value, h1, h2, prio)
                    return 1
            else:
                # Free slot
                if item_diff:
                    self.nitems += item_diff
                _node_set(tnode, key, value, h1, h2, prio)
                return 1

            # No room
            return 0

    def keys(self):
        # Allocate the list and fill it with CPython API to avoid the risk of invoking
        # the GC implicitly through calls to append
        cdef Py_ssize_t i, tsize, rvpos
        rv = PyList_New(self.nitems)
        rvpos = 0
        table = self.table
        tsize = self.table_size
        for i from 0 <= i < tsize:
            if table[i].key != NULL:
                assert rvpos < self.nitems
                Py_XINCREF(table[i].key)
                PyList_SET_ITEM(rv, rvpos, table[i].key)
                rvpos += 1
        assert rvpos == self.nitems

        # Leaking NULL entries would cause segfaults in optimized builds (no asserts)
        # So the following should be unreachable, but keep it here to avoid hard crashes
        while rvpos < self.nitems:
            Py_INCREF(None)
            PyList_SET_ITEM(rv, rvpos, <PyObject*>None)
        return rv

    def values(self):
        cdef Py_ssize_t i, tsize, rvpos
        rv = PyList_New(self.nitems)
        rvpos = 0
        table = self.table
        tsize = self.table_size
        for i from 0 <= i < tsize:
            if table[i].value != NULL:
                assert rvpos < self.nitems
                Py_XINCREF(table[i].value)
                PyList_SET_ITEM(rv, rvpos, table[i].value)
                rvpos += 1
        assert rvpos == self.nitems

        # Leaking NULL entries would cause segfaults in optimized builds (no asserts)
        # So the following should be unreachable, but keep it here to avoid hard crashes
        while rvpos < self.nitems:
            Py_INCREF(None)
            PyList_SET_ITEM(rv, rvpos, <PyObject*>None)
        return rv

    def items(self):
        # Only way to atomically snapshot this without risk of invoking the GC in the
        # middle due to python object allocations, is to build key and value lists
        # in one go, but in separate lists. We cannot build tuples on the go,
        # since each new allocation could trigger the GC and swap the table under
        # our feet.
        cdef Py_ssize_t i, tsize, rvkpos, rvvpos
        rvk = PyList_New(self.nitems)
        rvv = PyList_New(self.nitems)
        rvkpos = rvvpos = 0
        table = self.table
        tsize = self.table_size
        for i from 0 <= i < tsize:
            assert (table[i].key == NULL) == (table[i].value == NULL)
            if table[i].key != NULL:
                assert rvkpos < self.nitems
                Py_XINCREF(table[i].key)
                PyList_SET_ITEM(rvk, rvkpos, table[i].key)
                rvkpos += 1
            if table[i].value != NULL:
                assert rvvpos < self.nitems
                Py_XINCREF(table[i].value)
                PyList_SET_ITEM(rvv, rvvpos, table[i].value)
                rvvpos += 1
        assert rvkpos == self.nitems
        assert rvvpos == self.nitems

        # Leaking NULL entries would cause segfaults in optimized builds (no asserts)
        # So the following should be unreachable, but keep it here to avoid hard crashes
        while rvkpos < self.nitems:
            Py_INCREF(None)
            PyList_SET_ITEM(rvk, rvkpos, <PyObject*>None)
        while rvvpos < self.nitems:
            Py_INCREF(None)
            PyList_SET_ITEM(rvv, rvvpos, <PyObject*>None)
        return zip(rvk, rvv)

    def iterkeys(self):
        table = self.table
        tsize = self.table_size
        for i from 0 <= i < tsize:
            if table[i].key != NULL:
                k = <object>(table[i].key)
                yield k
                del k
                if self.table != table or self.table_size != tsize:
                    raise RuntimeError("LazyCuckooCache changed size during iteration")

    def itervalues(self):
        table = self.table
        tsize = self.table_size
        for i from 0 <= i < tsize:
            if table[i].value != NULL:
                k = <object>(table[i].value)
                yield k
                del k
                if self.table != table or self.table_size != tsize:
                    raise RuntimeError("LazyCuckooCache changed size during iteration")

    def iteritems(self):
        table = self.table
        tsize = self.table_size
        for i from 0 <= i < tsize:
            if table[i].key != NULL and table[i].value != NULL:
                k = <object>(table[i].key)
                v = <object>(table[i].value)
                yield (k,v)
                del k,v
                if self.table != table or self.table_size != tsize:
                    raise RuntimeError("LazyCuckooCache changed size during iteration")

    def __iter__(self):
        return self.iterkeys()

    @cython.cdivision(True)
    def __setitem__(self, key, value):
        cdef unsigned int h1, h2, tsize

        h1 = self._hash1(key)
        h2 = self._hash2(key)
        table = self.table
        tsize = self.table_size
        prio = self._assign_prio()
        if not self._add_node(table, tsize, NULL, h1, h2, <PyObject*>key, <PyObject*>value, prio, 1, True):
            if self._rehash():
                # Try agin
                table = self.table
                tsize = self.table_size
                if self._add_node(table, tsize, NULL, h1, h2, <PyObject*>key, <PyObject*>value, prio, 1, True):
                    return

            # No room, evict some entry, pick one of the two options randomly
            # NOTE: Don't invoke python code until all manipulations of the table
            #   are done, or baaad things may happen
            table = self.table
            tsize = self.table_size
            tnode1 = table + (h1 % tsize)
            tnode2 = table + (h2 % tsize)
            prio1 = tnode1.prio
            prio2 = tnode2.prio
            if prio1 < prio1 or (prio1 == prio2 and self._rnd()):
                tnode = tnode1
            else:
                tnode = tnode2
            eviction_callback = self.eviction_callback

            if tnode.key != NULL and tnode.value != NULL:
                k = <object>tnode.key
                v = <object>tnode.value
            else:
                # Should not happen, but better safe than segfault
                k = v = None
            _node_set(tnode, <PyObject*>key, <PyObject*>value, h1, h2, prio)

            # Notify eviction, atomic barrier
            if eviction_callback is not None:
                eviction_callback(k, v)

    @cython.cdivision(True)
    def __getitem__(self, key):
        cdef unsigned int tsize, h1, h2

        h1 = self._hash1(key)
        table = self.table
        tsize = self.table_size
        node = table + (h1 % tsize)
        if node.value != NULL:
            v = <object>node.value
            if _key_equals(node, key, h1):
                # Recheck table in case equals invoked re-entrant python code
                if self.touch_on_read and self.table == table:
                    node.prio = self._assign_prio()
                return v
            del v

        h2 = self._hash2(key)
        table = self.table
        tsize = self.table_size
        node = table + (h2 % tsize)
        if node.value != NULL:
            v = <object>node.value
            if _key_equals(node, key, h1):
                # Recheck table in case equals invoked re-entrant python code
                if self.touch_on_read and self.table == table:
                    node.prio = self._assign_prio()
                return v
            del v

        raise CacheMissError_(key)

    def __delitem__(self, key):
        self.pop(key)

    @cython.cdivision(True)
    def cas(self, key, oldvalue, newvalue):
        cdef unsigned int tsize, h1, h2
        cdef PyObject *tkey

        while 1:
            h1 = self._hash1(key)
            table = self.table
            tsize = self.table_size
            node = table + (h1 % tsize)
            tkey = node.key
            if _key_equals(node, key, h1):
                if self.table != table or node.key != tkey:
                    # Re-entrancy, restart operation
                    continue
                if node.value == <PyObject*>oldvalue:
                    _value_set(node, <PyObject*>newvalue, self._assign_prio())
                    return True
                else:
                    return False

            h2 = self._hash2(key)
            table = self.table
            tsize = self.table_size
            node = table + (h2 % tsize)
            tkey = node.key
            if _key_equals(node, key, h1):
                if self.table != table or node.key != tkey:
                    # Re-entrancy, restart operation
                    continue
                if node.value == <PyObject*>oldvalue:
                    _value_set(node, <PyObject*>newvalue, self._assign_prio())
                    return True
                else:
                    return False

            return False

    @cython.cdivision(True)
    cpdef get(self, key, deflt = None):
        cdef unsigned int tsize, h1, h2

        h1 = self._hash1(key)
        table = self.table
        tsize = self.table_size
        node = table + (h1 % tsize)
        if node.value != NULL:
            v = <object>node.value
            if _key_equals(node, key, h1):
                # Recheck table in case equals invoked re-entrant python code
                if self.touch_on_read and self.table == table:
                    node.prio = self._assign_prio()
                return v
            del v

        h2 = self._hash2(key)
        table = self.table
        tsize = self.table_size
        node = table + (h2 % tsize)
        if node.value != NULL:
            v = <object>node.value
            if _key_equals(node, key, h1):
                # Recheck table in case equals invoked re-entrant python code
                if self.touch_on_read and self.table == table:
                    node.prio = self._assign_prio()
                return v
            del v

        return deflt

    @cython.cdivision(True)
    cpdef pop(self, key, deflt = CacheMissError_):
        cdef unsigned int tsize, h1, h2
        cdef PyObject *tkey

        while 1:
            h1 = self._hash1(key)
            table = self.table
            tsize = self.table_size
            node = table + (h1 % tsize)
            tkey = node.key
            if _key_equals(node, key, h1):
                if self.table != table or tkey != node.key:
                    # Re-entrancy, retstart operation
                    continue
                if node.value != NULL:
                    rv = <object>node.value
                else:
                    rv = deflt
                _node_set(node, NULL, NULL, 0, 0, 0)
                self.nitems -= 1
                return rv

            h2 = self._hash2(key)
            table = self.table
            tsize = self.table_size
            node = table + (h2 % tsize)
            tkey = node.key
            if _key_equals(node, key, h1):
                if self.table != table or tkey != node.key:
                    # Re-entrancy, retstart operation
                    continue
                if node.value != NULL:
                    rv = <object>node.value
                else:
                    rv = deflt
                _node_set(node, NULL, NULL, 0, 0, 0)
                self.nitems -= 1
                return rv

            if deflt is CacheMissError_:
                raise CacheMissError_(key)
            else:
                return deflt

    @cython.cdivision(True)
    cpdef setdefault(self, key, deflt = None):
        cdef unsigned int tsize, h1, h2, j

        for j from 0 <= j <= 2:
            h1 = self._hash1(key)
            table = self.table
            tsize = self.table_size
            node1 = table + (h1 % tsize)
            if node1.value != NULL:
                v = <object>node1.value
                if _key_equals(node1, key, h1):
                    # Recheck table in case equals invoked re-entrant python code
                    if self.touch_on_read and self.table == table:
                        node1.prio = self._assign_prio()
                    return v
                del v

            h2 = self._hash2(key)
            table = self.table
            tsize = self.table_size
            node2 = table + (h2 % tsize)
            if node2.value != NULL:
                v = <object>node2.value
                if _key_equals(node2, key, h1):
                    # Recheck table in case equals invoked re-entrant python code
                    if self.touch_on_read and self.table == table:
                        node2.prio = self._assign_prio()
                    return v
                del v

            prio = self._assign_prio()
            if node1.key == NULL or node1.value == NULL:
                _node_set(node1, <PyObject*>key, <PyObject*>deflt, h1, h2, prio)
                self.nitems += 1
            elif node2.key == NULL or node2.value == NULL:
                _node_set(node2, <PyObject*>key, <PyObject*>deflt, h1, h2, prio)
                self.nitems += 1
            elif j < 2 and self._rehash():
                # Enlarged the table, retry
                continue
            else:
                # Pick a node to evict
                if self._rnd():
                    tnode = node1
                else:
                    tnode = node2
                eviction_callback = self.eviction_callback
                if tnode.key != NULL and tnode.value != NULL:
                    k = <object>tnode.key
                    v = <object>tnode.value
                else:
                    # Should not happen, but better safe than segfault
                    k = v = None
                _node_set(tnode, <PyObject*>key, <PyObject*>deflt, h1, h2, prio)
                if eviction_callback is not None:
                    eviction_callback(k, v)
            return deflt
        else:
            # Unreachable code, in theory
            raise AssertionError("unreachable")

    def update(self, iterOrDict):
        if self is iterOrDict:
            return
        if isinstance(iterOrDict, dict) or isinstance(iterOrDict, LazyCuckooCache):
            for k,v in iterOrDict.iteritems():
                self[k] = v
        else:
            for k,v in iterOrDict:
                self[k] = v

    def clear(self):
        cdef unsigned int tsize

        otable = self.table
        tsize = self.table_size
        self.table = _alloc_table(self.initial_size)
        self.table_size = self.initial_size
        self.nitems = 0
        _free_table(otable, tsize)

    def defrag(self):
        pass

    def __repr__(self):
        return "<LazyCuckooCache (%d elements, %d max)>" % (len(self), self.size)

# Cython's automatically generated tp_traverse and tp_clear don't know about our
# C-level hash table and all the references it contains.
#
# So we generate aditional traverse/clear methods and associate them to the
# type object at import time, remembering the original ones to chain them.
#
# A bit hackish, maybe, but effective, and Cython gives us no other recourse.

cdef traverseproc cuckoocache_cy_traverse = NULL
cdef inquiry cuckoocache_cy_clear = NULL

cdef int cuckoocache_traverse(PyObject *o, visitproc visit, void *arg):
    cdef LazyCuckooCache p = <LazyCuckooCache>o
    cdef int e
    cdef unsigned int i, tsize
    if cuckoocache_cy_traverse != NULL:
        e = cuckoocache_cy_traverse(o, visit, arg)
        if e:
            return e
    if p.table != NULL and p.table_size > 0:
        tsize = p.table_size
        table = p.table
        for i from 0 <= i < tsize:
            node = table + i
            if node.key != NULL:
                e = visit(node.key, arg)
                if e:
                    return e
            if node.value != NULL:
                e = visit(node.value, arg)
                if e:
                    return e
    return 0

cdef int cuckoocache_clear(o):
    cdef LazyCuckooCache p = <LazyCuckooCache>o
    cdef int e
    cdef unsigned int i, tsize
    if cuckoocache_cy_clear != NULL:
        e = cuckoocache_cy_clear(p)
        if e:
            return e
    if p.table != NULL and p.table_size > 0:
        _free_table_items(p.table, 0, p.table_size)

        # just to keep it consistent, shouldn't matter, but better safe than sorry
        p.nitems = 0
    return 0

cdef void lazy_cuckoocache_enable_gc(PyTypeObject *t):
    if t.tp_traverse != <traverseproc>cuckoocache_traverse:
        cuckoocache_cy_traverse = t.tp_traverse
        t.tp_traverse = <traverseproc>cuckoocache_traverse
    if t.tp_clear != <inquiry>cuckoocache_clear:
        t.tp_clear = <inquiry>cuckoocache_clear

lazy_cuckoocache_enable_gc(<PyTypeObject*>LazyCuckooCache)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import os
import sys
import multiprocessing
from multiprocessing.pool import ThreadPool
from setuptools.command.build_ext import build_ext
from Cython.Distutils import build_ext as cython_build_ext, Extension as CythonExtension
from Cython.Build import cythonize

from setuptools import setup, find_packages

cmd_class = {}
setup_dir = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(setup_dir, "README.rst")) as readme_file:
    readme = readme_file.read()

with open(os.path.join(setup_dir, "CHANGELOG.rst")) as history_file:
    history = history_file.read()

with open(os.path.join(setup_dir, "clru", "VERSION"), "r",) as vf:
    version = vf.read().strip()


def parse_requirements_txt(filename="requirements.txt"):
    with open(os.path.join(setup_dir, filename)) as requirements_file:
        requirements = requirements_file.readlines()
        # remove all the requirements that are comments
        requirements = [line for line in requirements if not line.startswith("#")]
        # remove inline comments
        requirements = [line.split("#", 1)[0] if "#" in line else line for line in requirements]
        # remove empty lines
        requirements = list(filter(None, requirements))
        # remove whitespaces
        requirements = [line.strip().replace(" ", "") for line in requirements]
        return requirements


def solve_transitive_dependencies(cython_extensions):
    """ Solves the dependencies based on the information of its dependencies.

    If file A depends on B, and B depends on C, then A also depends on C.

    :type cython_extensions: list(dict)
    :param cython_extensions: all the information of the extensions to create
        to cythonize the files. Each dict has the following keys:

        - name: the extension"s full package name
        - sources: a list with the python or cython file
        - depends: a list of dependencies. It can be the path to the PXD file
            or another name on the list. When possible we should add the
            referneces to this dict, so it can add the dependencies tree
            correctly
        - kwargs: a dict with the extra kwargs to use when building the
            Cython extension

    :rtype: list(:class:`.Extension`)
    :returns: all the extensions used to cythonize the files
    """
    # create a map with all the names of the extensions
    cython_extensions_map = {extension["name"]: extension for extension in cython_extensions}

    for extension in cython_extensions:
        extension["finished"] = False

    changed = True
    while changed:
        changed = False
        for extension in cython_extensions:
            finished = True
            final_dependencies = set()
            if extension["finished"]:
                # the dependencies are final
                continue

            for dependency in extension["depends"]:
                cython_dependency = cython_extensions_map.get(dependency, None)
                if cython_dependency is None:
                    # in this case is must reference a file
                    final_dependencies.add(dependency)
                    continue
                if not cython_dependency["finished"]:
                    finished = False
                    break

                final_dependencies.update(cython_dependency["depends"])

            if finished:
                # make sure that it is finished and all the dependencies
                # have been replaced for the corresponding files
                not_cython_files = [
                    filename
                    for filename in final_dependencies
                    if os.path.splitext(filename)[1] not in (".pxd", ".pyx", ".c", ".h")
                ]
                if not_cython_files:
                    # fmt: off
                    raise Exception(
                        "Extension %s was finished but a dependency isn't a cython file: %s" % (
                            extension["name"],
                            not_cython_files
                        )
                    )
                    # fmt: on

                extension["depends"] = list(final_dependencies)
                extension["finished"] = True
                changed = True

    # if all the values have been changed make sure that
    # all the extension are finished
    not_finished = [extension for extension in cython_extensions if not extension["finished"]]
    if not_finished:
        raise Exception("At least one extension hasn't finish: %s" % not_finished)

    extension_modules = [
        CythonExtension(
            extension["name"], extension["sources"], depends=extension["depends"], **extension.get("kwargs", {})
        )
        for extension in cython_extensions
    ]
    return extension_modules


# maps all the information of the files that should be cythonized and it"s
# dependencies. Check :meth:`.solve_transitive_dependencies` to get more information
# about the values on the dict. The depends should only include the cimported
# dependencies
cython_extensions = [
    {
        "name": "clru.lrucache.cylrucache",
        "sources": ["clru/lrucache/cylrucache.pyx"],
        "depends": ["clru/lrucache/cylrucache.pxd"],
        "kwargs": {"extra_compile_args": ["-O3"]},
    },
    {
        "name": "clru.cuckoocache.cycuckoocache",
        "sources": ["clru/cuckoocache/cycuckoocache.pyx"],
        "depends": ["clru/cuckoocache/cycuckoocache.pxd"],
        "kwargs": {"extra_compile_args": ["-O3"]},
    },
]
extension_modules = solve_transitive_dependencies(cython_extensions)
include_dirs = os.environ.get("CYTHON_INCLUDE_DIRS", ".").split(":")

force_cython = "--force-cython" in sys.argv
if force_cython:
    del sys.argv[sys.argv.index("--force-cython")]

parallel = None
if "-j" in sys.argv:
    jobpos = sys.argv.index("-j")
    if len(sys.argv) > jobpos:
        # check if the user specified the number of parallel jobs
        # or if we should use the number of CPU
        try:
            parallel = int(sys.argv[jobpos + 1])
            del sys.argv[jobpos : jobpos + 2]
        except ValueError as e:
            parallel = multiprocessing.cpu_count()
            del sys.argv[jobpos]

        if parallel == 1:
            parallel = None

if "--no-cython" in sys.argv:
    del sys.argv[sys.argv.index("--no-cython")]
    ext_modules = None
else:
    # check if Cython is already installed then we don"t have to do
    # anything special
    ext_modules = cythonize(extension_modules, include_path=include_dirs, nthreads=parallel, force=force_cython)
    cmd_class = {"build_ext": cython_build_ext}

    # use a coustom builder when doing a cythonization
    if parallel is not None:

        class parallel_build_ext(build_ext):
            def build_extensions(self):
                # First, sanity-check the "extensions" list
                self.check_extensions_list(self.extensions)

                for ext in self.extensions:
                    ext.sources = self.cython_sources(ext.sources, ext)

                pool = ThreadPool(parallel)
                pool.map(self.build_extension, self.extensions, chunksize=1)
                pool.close()
                pool.join()

        cmd_class["build_ext"] = parallel_build_ext


setup(
    name="clru",
    description="Cython LRU Structures",
    version=version,
    author="Claudio Freire",
    author_email="klaussfreire@gmail.com",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries",
        "Operating System :: OS Independent",
    ],
    install_requires=parse_requirements_txt(),
    extras_require={
        "dev": parse_requirements_txt("requirements-dev.txt"),
        "dev-strict": [req.replace(">=", "==") for req in parse_requirements_txt("requirements-dev.txt")],
    },
    license="LGPLv3",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    package_data={"": ["*.pxd", "*.pyx"], "clru": ["VERSION"]},
    packages=find_packages(include=["clru", "clru.*"]),
    test_suite="tests",
    tests_require=["pytest"],
    ext_modules=ext_modules,
    cmdclass=cmd_class,
    url="https://github.com/klaussfreire/clru",
    zip_safe=False,
)



"""Efficient (O(log n) amortized update) Least-Recently-Used cache"""

cdef extern from "Python.h":
    int PySequence_SetItem(object o, Py_ssize_t i, object v) except -1
    object PySequence_GetItem(object o, Py_ssize_t i)
    void PyList_SET_ITEM(void *o, Py_ssize_t i, void* v)
    void* PyList_GET_ITEM(void *o, Py_ssize_t i)
    Py_ssize_t PyList_GET_SIZE(void *o)

cdef extern from "lrucache_cy028_compat.h":
    pass

cdef extern from *:
    # Note: the C name below is extracted from the generated code. As such,
    #       it can change from time to time. It's a hack. Beware.
    #       the C above is a compatibility define for Cython 0.28 and above
    struct _borrowed_node "__pyx_obj_8lrucache__node":
        unsigned int prio
        unsigned int index

IsThreadsafe = True

CacheMissError = KeyError
cdef object CacheMissError_ = CacheMissError

cdef class _node:
    # attributes in pxd

    def __cinit__(_node self not None, unsigned int prio, unsigned int index, object key, object value):
        self.prio = prio
        self.index = index
        self.key = key
        self.value = value

    cdef attach(_node self, unsigned int prio, unsigned int index, object key, object value):
        self.prio = prio
        self.index = index
        self.key = key
        self.value = value

    cdef detach(_node self):
        self.key = self.value = None

    def __richcmp__(_node self not None, _node other, int op):
        if op == 0:
            return self.prio < other.prio
        elif op == 1:
            return self.prio <= other.prio
        elif op == 2:
            return self.prio == other.prio
        elif op == 3:
            return self.prio != other.prio
        elif op == 4:
            return self.prio > other.prio
        elif op == 5:
            return self.prio >= other.prio
        else:
            return False


cdef class LRUCache:
    """Least-Recently-Used (LRU) cache.
    
    Instances of this class behave like a Python mapping type, except that
    its size is guaranteed to never surpass the assigned limit and when
    spillover occurs, the least recently used items get removed first.

    If an eviction callback is provided, it must take two arguments, key and
    value, and it will only be invoked for automatic (overflow related) evictions,
    not manual ones invoked with del, pop or clear.
    """

    # attributes in pxd

    def __cinit__(LRUCache self, unsigned int size, unsigned int touch_on_read = 1, object eviction_callback = None,
            preallocate = True):
        cdef int i
        
        self.size = size
        self.touch_on_read = touch_on_read
        self.pqueue = []
        self.emap = {}
        self.next_prio = 0
        self.eviction_callback = eviction_callback

        if size > 0 and preallocate:
            # Preallocate big structures
            self.use_freelist = 1
            self.freelist = [None]*(size+1)
            for i from 0 <= i < size+1:
                self.freelist[i] = _node(0, 0, None, None)
            self.pqueue.extend(self.freelist)
            for i from 0 <= i < size+1:
                self.emap[i] = i
            self.emap.clear()
            del self.pqueue[:]
        else:
            self.freelist = []
            self.use_freelist = 0
    
    def __len__(LRUCache self not None):
        return len(self.pqueue)
    
    def __contains__(LRUCache self not None, object key):
        return key in self.emap

    cdef void c_rehash(LRUCache self):        
        cdef _borrowed_node *node
        cdef unsigned int i, sz
        cdef unsigned int bprio

        sz = <unsigned int>PyList_GET_SIZE(<void*>self.pqueue)
        if sz:
            bprio = (<_borrowed_node*>PyList_GET_ITEM(<void*>self.pqueue, 0)).prio
        else:
            bprio = self.next_prio
        for i from 0 <= i < sz:
            node = <_borrowed_node*>PyList_GET_ITEM(<void*>self.pqueue, i)
            node.prio = node.prio - bprio
        self.next_prio = self.next_prio - bprio

    cdef void c_decrease(LRUCache self, _node node):
        cdef unsigned int ix, l, r, sw, sz
        cdef _borrowed_node *ln, *rn, *swn, *bnode

        if self.next_prio > 0 and node.prio == self.next_prio - 1:
            # Nothing needs to be done, the node was recently decreased
            return

        node.prio = self.next_prio
        self.next_prio = self.next_prio + 1

        if self.next_prio == 0:
            self.c_rehash()

        # From here on, we work with borrowed references exclusively
        # This is possible since we're only shuffling items, in 
        # reference-neutral way, atomically, within a GIL-protected opcode
        bnode = <_borrowed_node*><void*>node
        sz = <unsigned int>PyList_GET_SIZE(<void*>self.pqueue)
        while 1:
            ix = bnode.index
            l  = 2 * ix + 1
            r  = 2 * ix + 2

            if r < sz:
                ln = <_borrowed_node*>PyList_GET_ITEM(<void*>self.pqueue,l)
                rn = <_borrowed_node*>PyList_GET_ITEM(<void*>self.pqueue,r)
            elif l < sz:
                ln = <_borrowed_node*>PyList_GET_ITEM(<void*>self.pqueue,l)
            else:
                break # just for safetey, should never happen

            if r < sz and rn.prio < ln.prio:
                sw = r
                swn= rn
            elif l < sz:
                sw = l
                swn= ln
            else:
                break

            # This is reference-neutral, so we can use the SET_ITEM macro
            PyList_SET_ITEM(<void*>self.pqueue, sw, <void*>bnode)
            PyList_SET_ITEM(<void*>self.pqueue, ix, <void*>swn)
            bnode.index = sw
            swn.index = ix

    def iterkeys(LRUCache self not None):
        return self.emap.iterkeys()

    def itervalues(LRUCache self not None):
        return LRUCacheValuesIterator(self)

    def iteritems(LRUCache self not None):
        return LRUCacheItemsIterator(self)

    def keys(LRUCache self not None):
        return self.emap.keys()

    def values(LRUCache self not None):
        return list(self.itervalues())

    def items(LRUCache self not None):
        return list(self.iteritems())

    def __iter__(LRUCache self not None):
        return self.iterkeys()

    cdef int c__setitem__(LRUCache self, object key, object val) except -1:
        cdef _node node
        cdef object oldkey, oldval
        cdef unsigned int cur_prio

        if key in self.emap:
            node = self.emap[key]
            # delay collection of old value, to avoid firing python code and thus releasing the GIL
            oldval = node.value 
            node.value = val
            self.c_decrease(node)
        elif <unsigned int>PyList_GET_SIZE(<void*>self.pqueue) >= self.size:
            node = <_node>PyList_GET_ITEM(<void*>self.pqueue, 0) # borrow pqueue, but new ref for node
            oldkey = node.key   # delay collection of old key/value, to avoid
            oldval = node.value # firing python code and thus releasing the GIL
            del self.emap[oldkey]
            node.key = key
            node.value = val
            self.emap[key] = node
            self.c_decrease(node)

            # Notify eviction, atomic barrier
            if self.eviction_callback is not None:
                self.eviction_callback(oldkey, oldval)
        else:
            cur_prio = self.next_prio
            self.next_prio = self.next_prio + 1
            if self.next_prio == 0:
                self.c_rehash()
            if self.use_freelist and len(self.freelist) > 0:
                node = self.freelist.pop()
                node.attach(self.next_prio, 0, key, val)
            else:
                node = _node.__new__(_node, self.next_prio, 0, key, val) # atomic barrier (might release GIL)
            node.index = <unsigned int>PyList_GET_SIZE(<void*>self.pqueue) # from now on, atomic
            self.pqueue.append(node)
            self.emap[key] = node
        return 0

    def __setitem__(LRUCache self not None, object key, object val):
        self.c__setitem__(key, val)

    cdef object c__getitem__(LRUCache self, key):
        cdef _node node

        node = self.emap.get(key)
        if node is None:
            raise CacheMissError_(key)
        else:
            if self.touch_on_read:
                self.c_decrease(node)
            return node.value
    
    def __getitem__(LRUCache self not None, key):
        return self.c__getitem__(key)

    cdef int c__delitem__(LRUCache self, key) except -1:
        cdef _node node, node2

        node = self.emap.get(key)
        if node is None:
            raise CacheMissError_(key)
        else:
            self.c_decrease(node)

            node2 = self.pqueue[-1]
            if node2 is not node:
                self.pqueue[node.index] = node2
                node2.index = node.index

            del self.emap[key]
            del self.pqueue[-1]

            if self.use_freelist and len(self.freelist) <= self.size:
                node.detach()
                self.freelist.append(node)
            
            return 0

    def __delitem__(LRUCache self not None, key):
        self.c__delitem__(key)

    def cas(LRUCache self not None, object key, object oldvalue, object newvalue):
        cdef _node node

        node = self.emap.get(key)
        if node is not None:
            if node.value is oldvalue:
                node.value = newvalue
                self.c_decrease(node)
                return True
            elif self.touch_on_read:
                self.c_decrease(node)
        return False

    cdef c_get(LRUCache self, object key, object deflt):
        cdef _node node

        node = self.emap.get(key)
        if node is None:
            return deflt
        else:
            self.c_decrease(node)
            return node.value
    
    def get(LRUCache self not None, object key, object deflt = None):
        return self.c_get(key, deflt)
    
    def pop(LRUCache self not None, object key, object deflt = CacheMissError_):
        cdef object rv

        if key not in self.emap:
            if deflt is CacheMissError_:
                raise CacheMissError_(key)
            else:
                rv = deflt
        else:
            rv = self.c__getitem__(key)
            self.c__delitem__(key)
        
        return rv

    def setdefault(LRUCache self not None, object key, object deflt = None):
        cdef _node node
        cdef object rv

        node = self.emap.get(key)
        if node is None:
            self.c__setitem__(key, deflt)
            return deflt
        else:
            rv = node.value
            if self.touch_on_read:
                self.c_decrease(node)
            return rv

    def update(LRUCache self not None, object iterOrDict):
        if isinstance(iterOrDict, dict) or isinstance(iterOrDict, LRUCache):
            for k,v in iterOrDict.iteritems():
                self[k] = v
        else:
            for k,v in iterOrDict:
                self[k] = v

    def clear(LRUCache self not None):
        # Hold onto old lists to prevent decref from freeing them before we're done
        cdef object pqueue, emap
        cdef object opqueue, oemap
        cdef object ofreelist, freelist
        opqueue = self.pqueue
        oemap = self.emap
        ofreelist = self.freelist
        pqueue = []
        freelist = []
        emap = {}
        self.pqueue = pqueue
        self.emap = emap
        self.freelist = freelist
        self.next_prio = 0
    
    def defrag(LRUCache self not None):
        # Hold onto old lists to prevent decref from freeing them before we're done
        cdef list pqueue
        cdef dict emap
        cdef object opqueue, oemap

        if self.use_freelist:
            # No need
            return
        
        opqueue = self.pqueue
        oemap = self.emap
        pqueue = list(opqueue)
        emap = oemap.copy()

        # Check consistency (previous copy was not atomic)
        if len(emap) != len(pqueue):
            return
        
        # swap
        self.pqueue = pqueue
        self.emap = emap

    def __repr__(LRUCache self not None):
        return "<LRUCache (%d elements, %d max)>" % (len(self), self.size)


cdef class LRUCacheItemsIterator:
    cdef LRUCache cache
    cdef unsigned int pos

    def __init__(LRUCacheItemsIterator self not None, LRUCache cache not None):
        self.cache = cache
        self.pos = 0
    def __next__(LRUCacheItemsIterator self not None):
        cdef _node node
        if self.pos >= len(self.cache.pqueue):
            raise StopIteration
        else:
            node = self.cache.pqueue[self.pos]
            self.pos = self.pos + 1
            return ( node.key, node.value )
    def __iter__(LRUCacheItemsIterator self not None):
        return self
    
cdef class LRUCacheValuesIterator:
    cdef LRUCache cache
    cdef unsigned int pos

    def __init__(LRUCacheValuesIterator self not None, LRUCache cache not None):
        self.cache = cache
        self.pos = 0
    def __next__(LRUCacheValuesIterator self not None):
        cdef _node node
        if self.pos >= len(self.cache.pqueue):
            raise StopIteration
        else:
            node = self.cache.pqueue[self.pos]
            self.pos = self.pos + 1
            return node.value
    def __iter__(LRUCacheValuesIterator self not None):
        return self
    
